use std::thread;

use godot::{classes::notify::NodeNotification, prelude::*};
use pipewire::{
    channel::{Receiver, Sender},
    context::Context,
    core::Core,
    main_loop::MainLoop,
    stream::Stream,
};

mod stream_node;

#[derive(GodotClass)]
#[class(base=Node)]
/// Node for interacting with Pipewire
/// Intended to not be spawned manually, as the plugin will create a
pub(crate) struct Pipewire {
    base: Base<Node>,

    /// Handle for the thread running the pipewire main loop
    thread_handle: Option<thread::JoinHandle<()>>,
    shutdown_sender: Sender<InboundPwSignal>,
}

#[derive(Debug, Clone, PartialEq)]
enum InboundPwSignal {
    Terminate,
    ConnectToStream(u32),
}

#[derive(Debug)]
enum ResultPwSignal {
    StreamConnection(Result<Stream, String>),
}

#[godot_api]
impl INode for Pipewire {
    fn init(base: Base<Node>) -> Self {
        let (thread, shutdown_sender) = Self::start_pipewire_thread();
        Self {
            base,
            thread_handle: Some(thread),
            shutdown_sender,
        }
    }
    fn on_notification(&mut self, what: NodeNotification) {
        if what == NodeNotification::PREDELETE
            || what == NodeNotification::UNPARENTED
            || what == NodeNotification::WM_CLOSE_REQUEST
            || what == NodeNotification::EXIT_TREE
        {
            godot_print!("Sending shutdown signal to pipewire thread");
            let _ = self.shutdown_sender.send(InboundPwSignal::Terminate);
            let handle = self.thread_handle.take().unwrap();
            let _join_result = handle.join();
            godot_print!("Pipewire thread completed");
        }
    }
    fn process(&mut self, _delta: f64) {}
}

#[godot_api]
impl Pipewire {
    pub const SINGLETON: &'static str = "Pipewire";

    fn start_pipewire_thread() -> (thread::JoinHandle<()>, Sender<InboundPwSignal>) {
        godot_print!("Starting pipewire thread");
        let (shutdown_sender, shutdown_receiver) = pipewire::channel::channel::<InboundPwSignal>();
        let thread_handle = thread::spawn(|| Self::pipewire_main(shutdown_receiver));
        (thread_handle, shutdown_sender)
    }

    /// Func that actually runs the main loop in the thread
    fn pipewire_main(shutdown_receiver: Receiver<InboundPwSignal>) {
        let mainloop = MainLoop::new(None).expect("Failed to create pipewire main loop");
        let context =
            Context::new(&mainloop).expect("Failed to get context from pipewire main loop");
        let core = context
            .connect(None)
            .expect("Failed to get core from pipewire context");

        let _receiver = shutdown_receiver.attach(mainloop.loop_(), {
            pipewire_loop_signal_handler(mainloop.clone(), core.clone())
        });

        godot_print!("Entering pipewire main loop");
        mainloop.run();
        println!("Pipewire main loop exited");
    }
}

fn pipewire_loop_signal_handler(mainloop: MainLoop, _core: Core) -> impl Fn(InboundPwSignal) {
    godot_print!("Stopping pipewire thread");
    move |signal| match signal {
        InboundPwSignal::Terminate => mainloop.quit(),
        InboundPwSignal::ConnectToStream(_stream_id) => {}
    }
}
